;;; pmwiki.el --- edit PmWiki pages within Emacs

;; Copyright (C) 2003  Christian Ridderstrm

;; Author:  Christian Ridderstrm <christian.ridderstrom@home.se>
;; Maintainer: Christian Ridderstrm <christian.ridderstrom@home.se>
;; Version: 1.0.0
;; Keywords: hypermedia
;; URL: http://www.emacswiki.org/cgi-bin/wiki.pl?PmWikiEditMode
;;
;;

;; This file is not part of GNU Emacs.

;;; License:

;; This is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; Background:
;;
;; This file is based on the elisp-files: simple-emacswiki.el,
;; simple-wiki-edit.el and simple-wiki.el, by
;;	Alex Schroeder <alex@gnu.org>
;; see http://www.emacswiki.org/cgi-bin/wiki.pl/SimpleWikiEditMode
;;
;; The files listed above were combined and partly rewritten into this
;; file, so that the functionality works with the WikiEngine PmWiki
;; (see	http://www.pmichaud.com/wiki)
;;
;;
;; Requirements:
;;
;; This file has been tested with Emacs 21.2.1 and requires these
;; additional packages:
;;	`http-post'	http://www.emacswiki.org/cgi-bin/wiki.pl/HttpPost
;;	`http-get'	http://www.emacswiki.org/cgi-bin/wiki.pl/HttpGet
;;
;; In addition, this package is useful:
;;	`longlines'	http://www.emacswiki.org/cgi-bin/wiki.pl/LongLines
;;
;; If you downloaded this file as a tar.gz-file, you should already
;; have these packages. Note that the version of `longlines' in the
;; archive has been patched - this line has been added to the end:
;;	(provide 'longlines)
;;
;;
;; Usage:
;;
;; Some examples to get you started:
;;
;; * Use the command `pmichaud-pmwiki-browse' to download a page from
;;   the PmWiki Wiki. Give 'Main' as the name of the group, and give
;;   'WikiSandbox' as the name of the page.
;;
;; * Use the command `pmwiki-open-url' to open a page by specifying
;;   its URL. For example, to open the page where this mode is
;;   described, give the following URL as argument to the command:
;;
;;	http://www.pmichaud.com/wiki/Cookbook/EmacsPmWikiMode
;;
;; * 
;;
;;
;; Just testing it...:
;;
;; A simple way to test this mode is to do as follows:
;;
;; 1. Add the directory, typically http-emacs, where this file is stored to
;;    the load-path of Emacs, e.g add the text.
;;
;;	(add-to-list 'load-path "~/public_html/Public/emacs/http-emacs")
;;
;;    to the file .emacs.
;;
;; 2. Restart emacs and use the command 'load-file' to load this file,
;;    i.e. simple-pmwiki.el
;;
;; 3. Now use the command 'pmwiki-wiki-browse' to open the page
;;    'WikiSandbox' from the 'Main' group.
;;
;;    Make some changes and save the page using the keyboard shortcut:
;;	C-c C-c
;;
;;    Use a browser to check what happened at
;;
;;    	http://www.pmichaud.com/wiki/Main/WikiSandbox.
;;
;; Enjoy. /Christian
;;
;;
;; Installation:
;;
;; * byte-compile the file?
;; 
;;
;;
;; Configuration options:
;;
;; Note that the code below is meant to be inserted in .emacs (or an
;; equivalent file).
;; 
;; If you want to use any of the configuration examples below, first
;; insert this line of code:
;;
;;	(defvar pmwiki-edit-mode-hooks)
;;
;; If your auto-fill-mode is usually on, you might want to turn it off:
;;
;;	(add-hook 'pmwiki-edit-mode-hooks 'turn-off-auto-fill)
;;	
;;
;; Also consider using the package longlines.el, for details see
;;   http://www.emacswiki.org/cgi-bin/wiki.pl/LongLines 
;;
;; 	(add-hook 'pmwiki-edit-mode-hooks
;;	  'pmwiki-delayed-hook-longlines-mode-on)
;;	(add-hook 'pmwiki-save-before-hooks 'longlines-mode-off)
;;
;; in which case, also customize `pmwiki-fill-column' back to 70.
;; Note that you may have to patch the file longlines.el by adding
;; 	(provide 'longlines)
;; to the end of it. As an alternative configuration, this code also
;; makes the `longlines-mode' show the hard newlines.
;;
;; 	(add-hook 'pmwiki-edit-mode-hooks
;;	  'pmwiki-delayed-hook-longlines-mode-showing-hard-newlines)
;;	(add-hook 'pmwiki-save-before-hooks 'longlines-mode-off)
;;
;;
;; Using pcomplete within the page: 
;;
;;	(add-hook 'pmwiki-edit-mode-hooks  'pcomplete-pmwiki-setup)
;;
;; and thereafter, use  C-/ to pcomplete pages at point and M-/ for
;; `dabbrev-completion'. Consider also (setq pcomplete-ignore-case t)
;; Note that I haven't tested this since it relies on a function,
;; `swc-pages-completion' that I don't have. (Maybe it's in XEmacs?)
;;

;;
;; Note: PmWiki supports using '\' as line continuation character,
;; i.e. if it is the the last character on the line the linebreak is
;; ignored. This is not supported.
;;

;;; Code:

;; Pattern of a group name (<GroupName>): AB..[c..] | A..b..(A..[b..])*
(defvar pmwiki-GroupName-pattern
  (concat 
   "\\("			; Two alternatives
   "\\(?:" "[A-Z]\\{2,\\}" "[a-z0-9]*" "\\)" 	 ; AB..[c..]
   "\\|"				     
   "\\(?:[A-Z]+[a-z0-9]+\\(?:[A-Z]+[a-z0-9]*\\)*\\)" ;A...b...(A...[b...])*
   "\\)" )
  "Pattern used to define a GroupName of a PmWiki-page")

;; Pattern of group-page separator: {. | /}
(defvar pmwiki-Group-Page-separator "\\(?:[./]\\)"
  "Pattern used to separate a pmwiki-GroupName-pattern from
a pmwiki-PageName-pattern" )

;; Pattern of page name (<PageName>): AB..c.. | A..b..(A..[b..])+
(defvar pmwiki-PageName-pattern
  (concat 
   "\\("			; Begin mandatory pattern for PageTitle
   "\\(?:" "[A-Z]+\\{2,\\}" "[a-z0-9]+" "\\)" "\\|"
   "\\(?:" "[A-Z]+" "[a-z0-9]+"  "\\(?:[A-Z]+" "[a-z0-9]*" "\\)+" "\\)"
   "\\)"			; End mandatory pattern for PageTitle
   )
  "Pattern that defines a PageName of a PmWiki-page.")

;; Pattern of WikiWiki-link: [<GroupName>{.|/}]<PageName>
(defvar pmwiki-link-pattern
  (concat 
   "\\<" 			; Force beginning of word?
   "\\(?:" pmwiki-GroupName-pattern pmwiki-Group-Page-separator "\\)?" 
   pmwiki-PageName-pattern
   )
  "The pattern used for finding WikiName.")

;; Have to modify patterns for faces. List of faces:
;; builtin, comment, constant, doc, function, keyword, string, type, variable
;; warning
;;	[=...=]			- string
;;	[[#<target>]]		- constant
;;	[[<url> <text>]]	- type
;;	{{link}}		- keyword
;;	Wiki-links		- keyword
;;	 <monospace>		- comment
;;	@@<monospace>@@		- comment
;;	''<emp>''		- type
;;	'''<strong>'''		- function
;;	'''''<extra>'''''	- warning
;;	*...			- builtin
;;	#...			- doc
;;	!...			- function
;;	:..: : <indented par.>	- doc
;;
;; Maybe remove sgml-tag-alist?
(define-derived-mode pmwiki-mode text-mode "PmWiki"
  "Simple mode to edit wiki pages.

\\{pmwiki-mode-map}"
  (font-lock-add-keywords
   nil
   '(
     ("^[ \t].+?$" . font-lock-comment-face)		;  <monospace>
     ("\\[=[^=]*=\\]" . font-lock-function-name-face)	; [= ... =]
     ("@@[^ \t@][^@]*[^ \t@]@@" . font-lock-comment-face) ; @@a .. b@@
     ("\\(^[*#]+\\)[^*#]" 1 font-lock-constant-face)	; {*|#}+ ...
     ("^!+[^!].*" . font-lock-constant-face)		; !+ ...
     ("\\[\\[<<\\]\\]" . font-lock-comment-face)	; [[<<]]
     ("\\[\\[#[^]]+\\]\\]" . font-lock-constant-face)	; [[#<target>]]
     ("\\\\$" . font-lock-keyword-face)			;     \
     ("\\\[\[[^#].*\]\]" . font-lock-type-face)		; [[ ... ]]
     ("{{[^}]*}}" . font-lock-keyword-face)		; {{ ... }}
     ("^:+ +:" . font-lock-doc-face)			; :..: : 
     ("^:[ \t]*[^ \t:]+[ \t]*+:" . font-lock-string-face)	; : term : 
     ("%[^ \t%][^%]*[^ \t%]%" . font-lock-warning-face)	; %color = red%
     ))
   (font-lock-add-keywords
   nil
   (list (cons (symbol-value 'pmwiki-link-pattern) 'font-lock-keyword-face))
   )

  (font-lock-mode 1)
  (goto-address)
  (set (make-local-variable 'sgml-tag-alist)
       `(("b") ("code") ("em") ("i") ("strong") ("nowiki")
	 ("pre" \n) ("tt") ("u")))
  (set (make-local-variable 'skeleton-transformation) 'identity)
  (setq indent-tabs-mode nil))


(require 'thingatpt)
(require 'http-get)

(defvar pmwiki-edit-version "1.0")

(defvar pmwiki-url nil
  "The URL of the current buffer.")

(defvar pmwiki-save-function nil
  "The function to use to save the current buffer.")

(defcustom pmwiki-fill-column 70 "Fill column width for `pmwiki-edit-mode'.")
(defcustom pmwiki-edit-mode-hooks nil "")
(defcustom pmwiki-save-before-hooks nil "")

(define-derived-mode pmwiki-edit-mode pmwiki-mode "PmWiki-Edit"
  "Edit URL using `pmwiki-mode'.

\\{pmwiki-edit-mode-map}"
  (run-hooks 'pmwiki-edit-mode-hooks)
  )

(define-key pmwiki-edit-mode-map (kbd "C-c C-c") 'pmwiki-save)
(define-key pmwiki-edit-mode-map (kbd "C-c C-o") 'pmwiki-open)
(define-key pmwiki-edit-mode-map (kbd "C-c C-g") 'pmwiki-follow)
(define-key pmwiki-edit-mode-map (kbd "C-c C-n") 'pmwiki-next)
(define-key pmwiki-edit-mode-map (kbd "C-c C-p") 'pmwiki-prev)
(define-key pmwiki-edit-mode-map (kbd "C-/") 'pcomplete)
(define-key pmwiki-edit-mode-map (kbd "C-c C-/") 'pcomplete)
(define-key pmwiki-edit-mode-map (kbd "C-c C-_") 'pcomplete)
(define-key pmwiki-edit-mode-map (kbd "C-c TAB") 'pcomplete)


;;;###autoload
(defun pmwiki-edit (url link &optional save-func http-ver content-type)
  "Edit URL using `pmwiki-edit-mode'.
Optional SAVE-FUNC is a function to use when saving."
  (unless content-type (setq content-type 'iso-8859-1))
  (let ((dst (concat (pmwiki-create-url-link link url) "?action=source"))
	(bufname (concat (car link) "." (cdr link)))
	(headers '(("Connection" "close"))))
    (http-get dst headers nil http-ver bufname content-type)
    (pmwiki-edit-mode)
    (set-fill-column pmwiki-fill-column)
    (set (make-local-variable 'pmwiki-url) dst)
    (set (make-local-variable 'pmwiki-save-function) save-func)))


(defun pmwiki-next ()
  "Goto the next WikiName."
  (interactive)
  (let ((case-fold-search nil))
    (re-search-forward
     pmwiki-link-pattern
     nil t)))

(defun pmwiki-prev ()
  "Goto the previous WikiName."
  (interactive)
  (let ((case-fold-search nil))
    (re-search-backward
     pmwiki-link-pattern
     nil t)))


(defun pmwiki-link-at-point ()
  "Return PmWiki-link in `cons' (\"<group>\", \"<page>\") at point 
of the current buffer. If no link is found at the current point, nil is
returned."
  (let ((case-fold-search nil)
	(found-full-link (thing-at-point-looking-at
			  (concat pmwiki-GroupName-pattern
				  pmwiki-Group-Page-separator
				  pmwiki-PageName-pattern))))
    (if found-full-link
	(cons (match-string 1) (match-string 2))
      (let ((found-link (thing-at-point-looking-at pmwiki-PageName-pattern))
	    (page (match-string 1)))
	(if found-link
	    (cons (pmwiki-url-part 2) page)
	  nil)))))


(defun pmwiki-follow ()
  "Follow the PmWiki-link at point."
  (interactive)
  (let ((link (pmwiki-link-at-point)))
    (if link (message (concat "Following: " (car link) "." (cdr link))))
    (if link
	(pmwiki-edit (pmwiki-url-part 1) link pmwiki-save-function)
      (error "No WikiName at point"))))



(defvar pmwiki-url-pattern
  (concat "^\\(.*\\)/"
	  pmwiki-GroupName-pattern "/"
	  pmwiki-PageName-pattern
	  "\\(?:\\?\\|$\\)" )
  "This pattern is used on `pmwiki-url' to extract various parts.
The parts matches \"<BaseURL>/<group>/<page>?\". where
 1 - <BaseURL>, 2 - <group> and 3 - <page>.
The parts of the pattern are derived from `pmwiki-GroupName-pattern'
and `pmwiki-PageName-pattern'.")

(defun pmwiki-url-part (part &optional url)
  "Extract BaseURL, GroupName or PageName from URL.
If URL is not given, `pmwiki-url' is used instead.
`pmwiki-url-pattern' is used to match to different parts of URL:
	URL = \"<BaseURL>/<group>/<page>?...\"
where PART determines what is returned according to this list:
	1 -> <BaseURL>
	2 -> <group>
	3 -> <page>"
  (unless url (set 'url pmwiki-url))
  (unless url (error "No URL given."))
  (let ((case-fold-search nil))
    (if (string-match pmwiki-url-pattern url)
	(match-string part url)
      (error "Cannot determine part %d in %s" part url))))
  

(defun pmwiki-create-url-link (link &optional base-url)
  "Create URL for a page specified by LINK,  with WikiEngine at BASE-URL.
The parameter link LINK is (cons <group-name> <page-name>).
The parameter BASE-URL is an optional URL to the WikiEngine and `pmwiki-url'
is used if BASE-URL is not given."
  (unless base-url (set 'base-url pmwiki-url))
  (unless base-url (error "No suitable BASE-URL available"))
  (if (string-match "\\(.*\\)/?$" base-url)	; Remove last '/'
      (concat (match-string 1 base-url) "/" (car link) "/" (cdr link))
    (error "Unable to remove last '/' from BASE-URL")))


(defun pmwiki-open (group page)
  "Open a new page from the same WikeEngine."
  (interactive "sGroup: \nsPage: ")
  (pmwiki-edit (pmwiki-url-part 1) (cons group page) pmwiki-save-function))


(defun pmwiki-open-url (url)
  "Open a new page specified by a URL."
  (interactive "sURL (ex: http://www.pmichaud.com/wiki/Main/WikiSandbox): ")
  (pmwiki-edit (pmwiki-url-part 1 url)
	       (cons (pmwiki-url-part 2 url) (pmwiki-url-part 3 url))
	       'pmwiki-save))


(defun pmwiki-save ()
  "Save the current buffer to the wiki."
  (interactive)
  (run-hooks 'pmwiki-save-before-hooks)
  (unless pmwiki-save-function
    (error "No save function specified"))
  (save-excursion
    (funcall pmwiki-save-function)))


;;
;; Functions useful for configuration.
;;
(defun pmwiki-delayed-hook-longlines-mode-on ()
  "Hook that activates `longlines-mode' when Emacs has been idle for 1 s."
  (require 'longlines)
  (run-with-idle-timer
   1 nil
   (lambda (buf)
     (save-excursion
       (set-buffer buf)
       (message (concat "Enabling longlines-mode for " (buffer-name)))
       (longlines-mode-on)))
   (current-buffer)))

(defun pmwiki-delayed-hook-longlines-mode-showing-hard-newlines ()
  "Hook that activates `longlines-mode' when Emacs has been idle for 1 s,
and in addition runs it `longlines-show-hard-newlines'."
  (require 'longlines)
  (run-with-idle-timer
   1 nil
   (lambda (buf)
     (save-excursion
       (set-buffer buf)
       (message (concat "Enabling longlines-mode for " (buffer-name)))
       (longlines-mode-on)
       (longlines-show-hard-newlines)))
   (current-buffer)))


(defun pcomplete-pmwiki-setup ()
  (set (make-local-variable 'pcomplete-parse-arguments-function)
   'pcomplete-parse-pmwiki-arguments)
  (set (make-local-variable 'pcomplete-default-completion-function)
       'pcomplete-pmwiki-default-completion)
  )


(defun pcomplete-parse-pmwiki-arguments ()
  (save-excursion
    (let* ((thispt (point))
	   (pt (search-backward-regexp "[ \t\n]" nil t))
	   (ptt (if pt (+ pt 1) thispt)))
      
      (list 
       (list "dummy" (buffer-substring-no-properties ptt thispt))
       (point-min) ptt))))


(defun pcomplete-pmwiki-default-completion ()
  (pcomplete-here swc-pages-completion))

;;; ************************************************************************
;;;
;;; Based on simple-emacswiki.el
;;;
(require 'http-post)

(defun pmwiki-wiki-save ()
  "Save the current page to a PmWiki wiki."
  (let ((cur-buffer (buffer-name))
	(url pmwiki-url)
	(pagename (concat (pmwiki-url-part 2) "." (pmwiki-url-part 3)))
	(save-func pmwiki-save-function))
    (message (concat "Output from http-post is in "
		   (buffer-name 
		    (process-buffer
		     (http-post
		      (pmwiki-url-part 1)	;; Get destination
		      (list (cons "pagename" pagename)
			    (cons "action" "post")
			    (cons "text" (buffer-string)))
		      'iso-8859-1)))))
    (switch-to-buffer cur-buffer)
    (pmwiki-edit-mode)
    (set (make-local-variable 'pmwiki-url) url)
    (set (make-local-variable 'pmwiki-save-function) save-func)
    ) 
  )

(defun pmichaud-pmwiki-browse (group page)
  "Browse the PmWiki Wiki. Ex: 'Main', 'WikiSandbox' will open the page from
	http://www.pmichaud.com/wiki/Main/WikiSandbox"
  (interactive "sGroup (ex: Main): \nsPage (ex: WikiSandbox): ")
  (pmwiki-edit "http://www.pmichaud.com/wiki"
	       (cons group page)
	       'pmwiki-wiki-save)
  )
  
(provide 'pmwiki)
;;; pmwiki.el ends here
